import blockwriter_base,blockreader_lseek
import os,mmap

class blockwriter_lseek(blockwriter_base.blockwriter_base):
    """
    See blockwriter_base documentation.
    This blockwriter stores blocks directly into a file, locating
    the correct offset with lseek. As we don't have (yet) 64-bits
    access in a clean way in Python, we have to do ugly hacks. Yuck.
    """
    def __init__(self,filename,blocksize,wholeblocks,lastblocksize):
        blockwriter_base.blockwriter_base.__init__(self,filename,blocksize,
                                                   wholeblocks,lastblocksize)
        self.filedesc = os.open (filename,os.O_RDWR|os.O_CREAT|os.O_TRUNC,0600)
        self.currentseek = 0 # current position (in blocks)
        self.maxseek = 2000000000/blocksize # max seekable offset (in blocks)
        self.locate(wholeblocks)
        try: os.lseek(self.filedesc,lastblocksize-1,1)
        except OSError: pass
        os.write (self.filedesc, "%c"%0)
        try: os.lseek(self.filedesc,-lastblocksize,1)
        except OSError: pass
    def _putblock(self,blocknumber,data):
        self.locate(blocknumber)
        os.write (self.filedesc,data)
        if len(data) != self.blocksize:
            try: os.lseek(self.filedesc,-len(data),1)
            except OSError: pass
        else: self.currentseek += 1
    # the "locate" method is crudely leeched from the lseek blockreader ! :-P
    locate = blockreader_lseek.blockreader_lseek.__dict__["locate"]



import blockwriter_base
import os,mmap

class blockwriter_mmap(blockwriter_base.blockwriter_base):
    """
    See blockwriter_base documentation.
    This blockwriter stores blocks directly into a memory-mapped file.
    The size of mapped files is limited by mmap itself.
    """
    def __init__(self,filename,blocksize,wholeblocks,lastblocksize):
        blockwriter_base.blockwriter_base.__init__(self,filename,blocksize,
                                                   wholeblocks,lastblocksize)
        self.filedesc    = os.open (self.filename,
                                    os.O_RDWR|os.O_CREAT|os.O_TRUNC,0600)
        self.filesize    = wholeblocks*blocksize+lastblocksize
        os.lseek (self.filedesc, self.filesize-1, 0) # 0 = SEEK_SET
        os.write (self.filedesc, "%c"%0)
        self.filemap = mmap.mmap (self.filedesc, self.filesize,
                                  prot=mmap.PROT_WRITE)

    def _putblock(self,blocknumber,data):
        begin = blocknumber*self.blocksize
        if blocknumber<self.wholeblocks: end=begin+self.blocksize
        else: end=self.filesize
        self.filemap[begin:end]=data



import blockwriter_base
import os,mmap

class blockwriter_stream(blockwriter_base.blockwriter_base):
    """
    See blockwriter_base documentation.
    This blockwriter stores blocks linearly, using a buffer. It outputs
    blocks as soon as possible, respecting block numbers.
    It allows "streaming", or on-the-fly handling of data.
    """
    def __init__(self,filename,blocksize,wholeblocks,lastblocksize):
        blockwriter_base.blockwriter_base.__init__(self,filename,blocksize,
                                                   wholeblocks,lastblocksize)
        self.filedesc    = os.open (self.filename,
                                    os.O_RDWR|os.O_CREAT|os.O_TRUNC,0600)
        self.buffers     = {}
        self.nextblock   = 0
    def _putblock(self,blocknumber,data):
        # check if we already had this block
        if blocknumber<self.nextblock:
            print "STREAM: block %d already written."%blocknumber
            return
        # queue block
        self.buffers[blocknumber]=data
        # check if there are blocks to dequeue
        while self.buffers.has_key(self.nextblock):
            os.write(self.filedesc,self.buffers[self.nextblock])
            del self.buffers[self.nextblock]
            self.nextblock += 1




#!/usr/bin/python2
import socket, string, struct, os, mmap, random, threading, time, sys
import ConfigParser
import blockwriter_mmap,blockwriter_lseek,blockwriter_stream

# utility function
def strlist(list):
    s = string.join(map (lambda (k,v):str(k),list[:10]))
    if len(list)>10:
        s += " (%d more)"%(len(list)-10)
    return s

# request announce
#if len(sys.argv)>1:
#    netsocket.sendto("FFFFFFFF",(sys.argv[1],2133))




class receiver(threading.Thread):
    def __init__(self,blockwriter,netsocket):
        threading.Thread.__init__(self)
        self.stamp = time.time()
        self.lock = threading.Lock()
        self.blockwriter = blockwriter
        self.lastpacketreceived = -1 # index of last received packet
        self.missing = {} # key=packetnumber, value=time of request
        self.lastpackettoreceive = self.blockwriter.blockcount-1
        self.netsocket = netsocket
        self.unicasttimeout = 2
        self.endofstreamdelay = 10
        self.blocksize = blockwriter.blocksize
        self.wholeblocks = blockwriter.wholeblocks
    def request(self,blocknumber):
        self.serveraddr = "193.55.63.165"
        self.serverport = 2133
        self.missing[blocknumber]=time.time()
        self.netsocket.sendto("%08X"%blocknumber,
                              (self.serveraddr,self.serverport))
    def finished(self):
        return \
               (not self.missing) and \
               (self.lastpacketreceived==self.lastpackettoreceive)
    def run(self):
        while not self.finished():
            (data,(addr,port))=self.netsocket.recvfrom\
                                (self.blockwriter.blocksize+20)
            self.addr = addr
            self.port = port

            # random packet loss !
            if 0==random.choice(range(0,10)): continue
    
            (blocknumber,data) = string.split (data, ":", 1)
            if blocknumber=="FFFFFFFF": continue
            self.stamp = time.time()
            blocknumber = string.atoi("0x"+blocknumber,16)
            if blocknumber<=self.lastpacketreceived:
                if blocknumber not in self.missing.keys():
                    print "I don't need block %d"%blocknumber
                    continue
                del self.missing[blocknumber]
            for b in range(self.lastpacketreceived+1,blocknumber):
                self.missing[b]=time.time()
            if blocknumber>self.lastpacketreceived:
                self.lastpacketreceived=blocknumber
            begin = blocknumber*self.blocksize
            if blocknumber < self.wholeblocks: end=begin+self.blocksize
            else: end=begin+self.blockwriter.lastblock
            if end-begin != len(data):
                print "ERROR: wrong packet size, expected=%d, got=%d"%\
                      (end-begin, len(data))
            self.blockwriter.putblock(blocknumber,data)
        print "Receiver thread exitting"

    def checkqueue(self):
        timedoutunicasts = filter (lambda (k,v),
                                   limit=time.time()-self.unicasttimeout:
                                   v<limit, self.missing.items())
        if len(timedoutunicasts)>0:
            print "%d unicast requests have timed out, requesting %s"%\
                  (len(timedoutunicasts), strlist(timedoutunicasts))
            for (b,t) in timedoutunicasts:
                self.request (b)
        if self.stamp+self.endofstreamdelay<time.time():
            print "no packet received since a long time, requesting %s"%\
                  strlist(missing.items())
            for b in self.missing.keys():
                self.request (b)
            if lastpacketreceived!=lastpackettoreceive:
                print "we lack the end of the stream, requesting %d packets"%\
                      lastpackettoreceive-lastpacketreceived
                for b in range(lastpacketreceived+1,lastpackettoreceive+1):
                    self.request (b)
    
class updater(threading.Thread):
    def __init__(self,receiver,syncinterval,checkqueueinterval,sleepinterval):
        threading.Thread.__init__(self)
        self.receiver = receiver
        self.syncinterval = syncinterval
        self.checkqueueinterval = checkqueueinterval
        self.sleepinterval = sleepinterval
    def run(self):
        self.nextsync = time.time()
        self.nextcheckqueue = time.time()
        while not self.receiver.finished():
            now = time.time ()
            if self.nextsync < now:
                os.system("sync")
                self.nextsync += self.syncinterval
            if self.nextcheckqueue < now:
                self.receiver.checkqueue()
                self.nextcheckqueue += self.checkqueueinterval
            time.sleep(self.sleepinterval)
        print "Updater thread exitting"

class wrapper:
    def __init__(self):
        mcastaddr  = "224.2.13.3"
        mcastport  = 2133
        netsocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        netsocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        netsocket.bind(('', mcastport))
        bytes = map(int, string.split(mcastaddr, "."))
        grpaddr = 0
        for byte in bytes: grpaddr = (grpaddr << 8) | byte
        ifaddr = socket.INADDR_ANY
        mreq = struct.pack('ll', socket.htonl(grpaddr), socket.htonl(ifaddr))
        netsocket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

        netsocket.sendto("FFFFFFFF",("193.55.63.165",2133))

        # first state : waiting for "announce" packet
        while 1:
            (data,(addr,port)) = netsocket.recvfrom(1600)
            data = string.split(data,":")
            if data[0]!="FFFFFFFF":
                continue
            params = {}
            for kv in data[1:]:
                (k,v) = string.split(kv,"=")
                params[k]=v
            break

        # second state : send registration to the server
        netsocket.sendto("FFFFFFFF",(addr,port))

        # receive data
        filename = params['filename']
        blocksize = int(params['blocksize'])
        lastblock = int(params['lastblock'])
        wholeblocks = int(params['wholeblocks'])
        print "receiving: ",filename
        blockwriter = blockwriter_stream.blockwriter_stream("RECV-%s"%filename,
                                                            blocksize,
                                                            wholeblocks,
                                                            lastblock)
        r = receiver (blockwriter,netsocket)
        r.start()
        u = updater (r,1,1,1)
        u.start()

        while r.missing or \
              r.lastpackettoreceive!=r.lastpacketreceived:
            time.sleep(1)
            print "%d blocks left, %d missing : %s"%\
                  (r.lastpackettoreceive-r.lastpacketreceived,
                   len(r.missing),strlist(r.missing.items()))


w=wrapper()



import Tkinter, ConfigParser, string, time
cp = ConfigParser.ConfigParser ()
cp.read("example.dpk")

def display_name(s): return s
def display_comment(s): return s
def display_bitrate(s): return "%d mbps"%(int(s)/1000000)
def display_start(s):
    delay = int(s) - time.time()
    if delay<0: suffix=" ago"
    else: suffix=""
    delay=abs(delay)
    ddd=[ (60,"s",1), (60*60,"m",60), (60*60*24,"h",60*60),
          (1000000000, "d",60*60*24) ]
    for maxdelay,unit,divisor in ddd:
        if delay<maxdelay:
            return "%d%s%s"%(delay/divisor,unit,suffix)

class Flowtable(Tkinter.Frame):
    def __init__(self,root):
        Tkinter.Frame.__init__(self,root)
        self.root=root
        self.columns = ["DL","Name","Comment","Bitrate", "Start"]
        self.labels = {}
        for c in self.columns:
            self.labels[c] = Tkinter.Label(self.root,text=c)
            self.labels[c].grid(column=self.columns.index(c),row=0)
        self.rowcount = 0
        self.rows = {}
    def add(self,**row):
        self.rowcount += 1
        self.rows[self.rowcount]={}
        for c in self.columns:
            lc=string.lower(c)
            if c=="DL":
                self.rows[self.rowcount][c] = Tkinter.Checkbutton ()
                self.rows[self.rowcount][c].grid(column=self.columns.index(c),
                                                 row=self.rowcount)
                continue
            self.rows[self.rowcount][c] = Tkinter.Label(self.root,
                                                        text=globals()\
                                                        ["display_%s"%lc]\
                                                        (row[lc]))
            self.rows[self.rowcount][c].grid(column=self.columns.index(c),
                                             row=self.rowcount)
            
        
r=Tkinter.Tk()
ft=Flowtable(r)
for flowname in cp.sections():
    parms={}
    for p in cp.options(flowname):
        parms[p]=cp.get(flowname,p)
    ft.add(name=flowname,**parms)



import blockreader_base
import os,mmap

class blockreader_mmap(blockreader_base.blockreader_base):
    """
    See blockreader_base documentation.
    This blockreader maps a file into memory, using mmap.
    The size of mapped files is limited by mmap itself.
    """
    def __init__(self,filename,blocksize):
        self.filename    = filename
        self.filedesc    = os.open (self.filename,os.O_RDONLY)
        filesize         = os.stat(self.filename)[6]
        blockreader_base.blockreader_base.__init__(self,filesize,blocksize)
        self.filemap     = mmap.mmap (self.filedesc, self.filesize,
                                      prot=mmap.PROT_READ)
    def getblock(self,blocknumber):
        self.checkblock(blocknumber)
        begin = blocknumber*self.blocksize
        if blocknumber<self.filesize/self.blocksize: end=begin+self.blocksize
        else: end=self.filesize
        return self.filemap[begin:end]



import blockreader_base
import os

class blockreader_lseek(blockreader_base.blockreader_base):
    """
    See blockreader_base documentation.
    This blockreader uses os.lseek to allow access to files larger
    than 2GB. As Python can't handle 64 bits integers easily, we have
    to use *ugly* hacks.
    """
    def __init__(self,filename,blocksize):
        blockreader_base.blockreader_base.__init__(self,0,blocksize)
        self.filename = filename
        self.filedesc = os.open(self.filename,os.O_RDONLY)
        # first, go to the end of the file
        try: filesize = os.lseek(self.filedesc,0,2)
        except OSError: pass
        # then, try to count how many blocks we have
        # we lseek negatively until offset is less than blocksize,
        # ignoring errors (error means "offset larger than 2 GB")
        blockcount = 0
        while 1:
            try:
                blockcount += 1
                lastblock = os.lseek(self.filedesc,-blocksize,1)
                if lastblock<blocksize: break
            except OSError: pass
        self.blockcount  = blockcount
        self.lastblock   = lastblock
        self.wholeblocks = blockcount
        if self.lastblock: self.blockcount += 1
        # set some internal pointers
        os.lseek(self.filedesc,0,0)
        self.currentseek = 0 # current position (in blocks)
        self.maxseek = 2000000000/blocksize # max seekable offset (in blocks)
    def getblock(self,blocknumber):
        self.checkblock(blocknumber)
        self.locate(blocknumber)
        block = os.read(self.filedesc,self.blocksize)
        if len(block) != self.blocksize:
            try: os.lseek(self.filedesc,-len(block),1)
            except OSError: pass
        else: self.currentseek += 1
        return block
    def locate(self,blocknumber):
        while blocknumber != self.currentseek:
            remainingoffset = blocknumber - self.currentseek
            sign = abs(remainingoffset)/remainingoffset
            quantity = abs(remainingoffset)
            if quantity > self.maxseek: quantity = self.maxseek
            self.currentseek += sign*quantity
            try: os.lseek(self.filedesc,sign*quantity*self.blocksize,1)
            except OSError: pass



class blockreader_base:
    """
    Base class for other blockreaders. Each blockreader should
    have a constructor allowing to specify the file or data flow
    to read, and a blocksize. The blocksize might be automatically
    calculated according to other parameters, but it has to be
    known when the flow starts.

    This class allow automatic computation of the number of blocks
    according to blocksize and filesize. It also features a getnextblock
    method using a nextblock counter, for easy implementation of flow send.
    getnextblock will return None if the last block has been reached.

    Every subclass should define a getblock(blocknumber) method,
    returning the given block as a string. getblock methods can call
    checkblock, which ensures that the block number is valid.
    Requesting an invalid blocknumber (negative, or too large)
    should throw an exception (IndexError is appropriate).

    The rationale for using block numbers instead of byte offsets is
    that we can thus bypass the 4 Gb limit due to 32 bits offsets.
    """
    def __init__(self,filesize,blocksize):
        self.blocksize   = blocksize
        self.filesize    = filesize
        self.wholeblocks = self.filesize / self.blocksize
        self.lastblock   = self.filesize % self.blocksize
        self.blockcount  = self.wholeblocks
        if self.filesize % self.blocksize : self.blockcount += 1
        self.nextblock   = 0
    def getnextblock(self):
        if self.nextblock==self.blockcount: return None
        block = self.getblock(self.nextblock)
        self.nextblock += 1
        return block
    def checkblock(self,blocknumber):
        if blocknumber<0 or blocknumber>=self.blockcount:
            raise IndexError("block %d not in 0..%d"%\
                             (blocknumber,self.blockcount))
    def getblock(self,blocknumber): pass
        



class blockwriter_base:
    """
    Base class for other blockwriters. Each blockwriter should
    have a constructor with at least the same parameters as this
    one.
    
    Every subclass should define a _putblock(blocknumber,data) method,
    which should store data to the file or write it to the stream
    with appropriate buffering (that is, in the correct order).
    Calling programs will use the putblock method, which will make
    various checks, and call _putblock eventually.
    Requesting an invalid blocknumber (negative, or too large)
    will throw an IndexError ; trying to store a block of the wrong
    size will throw a ValueError.

    The rationale for using block numbers instead of byte offsets is
    that we can thus bypass the 4 Gb limit due to 32 bits offsets.
    """
    def __init__(self,filename,blocksize,wholeblocks,lastblocksize):
        self.filename      = filename
        self.blocksize     = blocksize     # size of all blocks, but last one
        self.wholeblocks   = wholeblocks   # number of blocks of this size
        self.lastblocksize = lastblocksize # size of last block (0 means same)
        self.blockcount    = wholeblocks   # total number of blocks
        if lastblock: self.blockcount += 1
    def checkblock(self,blocknumber,data):
        if blocknumber<0 or blocknumber>=self.blockcount:
            raise IndexError("block %d not in 0..%d"%\
                             (blocknumber,self.blockcount))
        if blocknumber<self.wholeblocks-1 and len(data)!=self.blocksize:
            raise ValueError("block %d has wrong size (%d instead of %d)"%\
                             (blocknumber,len(data),self.blocksize))
        if blocknumber==self.wholeblocks and len(data)!=self.lastblocksize:
            raise ValueError("block %d has wrong size (%d instead of %d)"%\
                             (blocknumber,len(data),self.lastblocksize))
    def putblock(self,blocknumber,data):
        self.checkblock(blocknumber,data)
        self._putblock(blocknumber,data)
    def _putblock(self,blocknumber,data): pass
        



#!/usr/bin/python2

import sys, socket, errno, os, struct, threading, string, time, mmap, select
import ConfigParser
import blockreader_mmap,blockreader_lseek

defaults = {
    "blockreader":         ("mmap",       str),
    "blocksize":           ("1000",       int),
    "multicastpacketrate": ("128",        int),
    "unicastpacketrate":   ("128",        int),
    "multicastgroup":      ("224.1.2.3",  str),
    "multicastport":       ("1234",       int),
    "multicastttl":        ("1",          int),
    "unicastaddr":         ("",           str),
    "unicastport":         ("4321",       int),
    "retransdelay":        ("3",          int),
    "retransmcast":        ("2",          int)
    }

def sleep(t):
    if (t<0): return
    now = time.time()+t
    while time.time() < now: pass

class stats:
    def __init__(self):
        self.unicasts    = 0   # unicast packets sent
        self.multicasts  = 0   # multicast packets sent
        self.clients     = {}  # retrans requests per (IP,PORT) tuple
        self.refills     = 0   # number of times the flow has started
    def __call__(self, unicast=0, multicast=0, client=None, refill=0):
        self.unicasts   += unicast
        self.multicasts += multicast
        self.refills    += refill
        if client:
            if not self.clients.has_key(client):
                self.clients[client]=0
            self.clients[client] = self.clients[client]+1
    def __repr__(self):
        return "U:%d M:%d R:%d dC:%d tC:%d"%\
               (self.unicasts,self.multicasts,self.refills,len(self.clients),
                reduce(lambda a,b:a+b,self.clients.values(),0))

class requestqueue:
    """
    Queue of packets to be re-sent. Packets can be added using
    push(), specifying the packet number and the client requesting the
    packet. If the packet is not queued yet, its queuing time is
    recorded. In all cases, the client is added to the list of clients
    having requested the packet.
    When calling pop(), the earliest packet whose deadline is anterior
    to specified time is dequeued and returned as a tuple :
    (packetnumber, list of clients).
    If there is no packet (empty queue, or no deadline reached yet),
    pop() returns None.
    """
    class request:
        def __init__(self,number):
            self.number  = number
            self.when    = time.time()
            self.clients = []
        def __repr__(self):
            return "(%d@%f:%s)"%(self.number,self.when,str(self.clients))
    def __init__(self):
        self.queue = []
        self.index = {}
        self.lock = threading.Lock()
    def push(self,blocknumber,client):
        self.lock.acquire()
        if not self.index.has_key(blocknumber):
            self.index[blocknumber] = requestqueue.request(blocknumber)
            self.queue += [self.index[blocknumber]]
        if client not in self.index[blocknumber].clients:
            self.index[blocknumber].clients += [ client ]
        self.lock.release()
    def pop(self,when):
        if len(self.queue)==0: return None
        if self.queue[0].when > when: return None
        else:
            self.lock.acquire()
            p = self.queue[0]
            self.queue = self.queue[1:]
            del self.index[p.number]
            self.lock.release()
            return (p.number, p.clients)

class receiver(threading.Thread):
    """
    Thread handling all incoming packets. It will receive retrans requests,
    and add them to the requestqueue. These requests will be handled later
    by the "sender" thread. The receiver thread also handles "info requests"
    (magic packet number 0xFFFFFFFF).
    """
    def __init__(self,filer):
        threading.Thread.__init__(self)
        self.filer = filer
    def run(self):
        self.running = Ellipsis
        while self.running:
            (a,b,c)=select.select ([self.filer.netsocket], [], [], 0)
            if a==[]:
                time.sleep(0.1)
                continue
            (data,(addr,port)) = self.filer.netsocket.recvfrom(1000)
            self.filer.stats(client=addr)
            data = string.split(data,":")
            if data[0]=="FFFFFFFF":
                #self.filer.announce(addr,port)
                self.filer.announce()
                if self.filer.nextmulticastblock == -1:
                    self.filer.nextmulticastblock = 0
                    self.filer.stats(refill=1)
            else:
                offset = string.atoi ("0x"+data[0],16)
                self.filer.requestqueue.push(offset,(addr,port))
        print "Receiver thread shutting down."

class sender(threading.Thread):
    """
    Thread handling main multicast flow and retransmissions. As this thread
    controls all outgoing traffic, it is able to limit unicast+multicast
    bandwidth if necessary.
    """
    def __init__(self,filer):
        self.filer = filer
        threading.Thread.__init__(self)
    def run(self):
        self.running = Ellipsis
        #lastannounce = time.time()
        lastannounce = 2000000000
        while self.running:
            if time.time()-lastannounce>1:
                self.filer.announce()
                lastannounce = time.time()
            while 1:
                pop = self.filer.requestqueue.pop(time.time()-
                                                  self.filer.retransdelay)
                if pop==None: break
                (blocknumber,clients) = pop
                if len(clients) < self.filer.retransmcast:
                    for (addr,port) in clients:
                        self.filer.sendblock(blocknumber,addr,port)
                else:
                    self.filer.sendblock(blocknumber)
            if self.filer.nextmulticastblock != -1:
                self.filer.sendblock(self.filer.nextmulticastblock)
                self.filer.nextmulticastblock += 1
                if self.filer.nextmulticastblock == \
                   self.filer.blockreader.blockcount:
                    self.filer.nextmulticastblock = -1
            else:
                time.sleep(0.01) # yield scheduler
            sleep (self.filer.interval)
        print "Sender thread shutting down."

class filer:
    def __init__(self,blockreader,blocksize,
                 unicastaddr,unicastport,
                 multicastgroup,multicastport,multicastttl,
                 multicastpacketrate,unicastpacketrate,
                 retransdelay,retransmcast,
                 filename):
        self.blockreader         = blockreader
        self.unicastaddr         = unicastaddr
        self.unicastport         = unicastport
        self.multicastgroup      = multicastgroup
        self.multicastport       = multicastport
        self.multicastttl        = multicastttl
        self.multicastpacketrate = multicastpacketrate
        self.unicastpacketrate   = unicastpacketrate
        self.retransdelay        = retransdelay
        self.retransmcast        = retransmcast
        self.netsocket   = socket.socket (socket.AF_INET, socket.SOCK_DGRAM)
        self.netsocket.bind ((unicastaddr, unicastport))
        #ttl = struct.pack('b',1)
        #netsocket.setsockopt (socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)

        self.interval = 1.0/multicastpacketrate

        self.sender      = sender (self)
        self.receiver    = receiver (self)
        self.stats       = stats()
        self.nextmulticastblock = -1
        self.requestqueue = requestqueue()

    def announce(self,addr=None,port=None):
        if addr==None: addr=self.multicastgroup
        if port==None: port=self.multicastport
        self.netsocket.sendto (
            ("FFFFFFFF:filename=%s:blocksize=%d:wholeblocks=%d:lastblock=%d"%\
             (self.blockreader.filename,self.blockreader.blocksize,
              self.blockreader.wholeblocks,self.blockreader.lastblock)),
            (addr,port))

    def sendblock(self,blocknumber,addr=None,port=None):
        if addr==None:
            addr=self.multicastgroup
            port=self.multicastport
            self.stats(multicast=1)
        else: self.stats(unicast=1)
        blockdata = self.blockreader.getblock(blocknumber)
        self.netsocket.sendto("%08X:%s"%(blocknumber,blockdata),(addr,port))

    def start(self):
        self.sender.start ()
        self.receiver.start ()

    def stop(self):
        self.sender.running = None
        self.receiver.running = None

class wrapper(filer):
    def __init__(self,configfile="deplika.conf",**extraopts):
        global defaults
        cp = ConfigParser.ConfigParser ()
        cp.read(configfile)
        options = {}
        for o in defaults.keys(): options[o]=defaults[o][0]
        for o in cp.options('bmcast'): options[o]=cp.get('bmcast',o)
        for o in extraopts.keys(): options[o]=extraopts[o]
        for o in defaults.keys(): options[o]=defaults[o][1](options[o])
        options['blockreader'] = globals()\
                                 ['blockreader_%s'%options['blockreader']].\
                                 __dict__\
                                 ['blockreader_%s'%options['blockreader']]\
                                 (options['filename'],options['blocksize'])
        filer.__init__(self,**options)
        
if sys.argv[0]!='':
    try:
        f = wrapper(filename=sys.argv[1])
        f.start()
        while 1:
            time.sleep(1)
            print f.stats
    except IndexError: print "Specify filename"
    except KeyboardInterrupt: f.stop()

